from ._abstract import (
    Quantity,
    Object,
)
from ._attribute import (
    Attribute,
)
from ._exceptions import (
    MetadataTypeError,
    MetadataValueError,
    OperationError,
)
from . import methods
from . import mixins
from ._operand import (
    Operand,
)
from ._operations import (
    unary,
    equality,
    ordering,
    additive,
    multiplicative,
)


__all__ = [
    # Modules
    methods,
    mixins,
    # Object classes
    Attribute,
    Quantity,
    Object,
    Operand,
    # Error classes
    MetadataTypeError,
    MetadataValueError,
    OperationError,
    # Functions
    additive,
    equality,
    multiplicative,
    ordering,
    unary,
]




class MetadataTypeError(TypeError):
    """A metadata-related TypeError occurred."""


class MetadataValueError(ValueError):
    """A metadata-related ValueError occurred."""


class OperationError(NotImplementedError):
    """A metadata attribute does not support this operation.
    
    The default behavior when applying an operator to a metadata attribute of
    `~Operand` is to copy the current value if the attribute does not define the
    operation. Custom metadata attributes may raise this exception to indicate
    that attempting to apply the operator is truly an error.
    """




from numerical import operators

from ._abstract import (
    Quantity,
    Object,
)
from ._exceptions import (
    MetadataTypeError,
    MetadataValueError,
)


def _build_error_message(
    f: operators.Operator,
    *args,
    error: str | None = None,
    key: str | None = None,
) -> str:
    """Helper for metadata error messages.

    This function should avoid raising an exception if at all possible, and
    instead return the default error message, since it is already being called
    as the result of an error elsewhere.
    """
    errmsg = f"Cannot compute {f}"
    errstr = error.lower() if isinstance(error, str) else ''
    if errstr == 'unequal':
        return f"{errmsg} between objects with unequal metadata"
    types = [type(arg) for arg in args]
    if errstr in {'non-empty', 'nonempty'}:
        if len(types) == 2:
            a, b = types
            endstr = "because {} has metadata"
            if issubclass(a, Object):
                return f"{errmsg} between {a} and {b} {endstr.format(str(a))}"
            if issubclass(b, Object):
                return f"{errmsg} between {a} and {b} {endstr.format(str(b))}"
    if errstr == 'type':
        if key is None:
            keystr = "a metadata attribute"
        else:
            keystr = f"metadata attribute {key!r}"
        midstr = f"because {keystr}"
        endstr = "does not support this operation"
        if len(types) == 1:
            return f"{errmsg} of {types[0]} {midstr} {endstr}"
        if len(types) == 2:
            a, b = types
            return f"{errmsg} between {a} and {b} {midstr} {endstr}"
    return errmsg


def unary(f: operators.Operator, a):
    """Compute the unary operation f(a)."""
    if isinstance(a, Quantity):
        meta = {}
        for key, value in a._meta.items():
            try:
                v = f(value)
            except TypeError as exc:
                errmsg = _build_error_message(f, a, error='type', key=key)
                raise MetadataTypeError(errmsg) from exc
            else:
                meta[key] = v
        return type(a)(f(a._data), **meta)
    return f(a)


def equality(f: operators.Operator, a, b):
    """Compute the equality operation f(a, b)."""
    x = a._data if isinstance(a, Quantity) else a
    y = b._data if isinstance(b, Quantity) else b
    fxy = f(x, y)
    isne = f(1, 2)
    try:
        iter(fxy)
    except TypeError:
        r = bool(fxy)
    else:
        r = all(fxy) or isne
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        if a._meta != b._meta:
            return isne
        return r
    if isinstance(a, Quantity):
        if not a._meta:
            return r
        return isne
    if isinstance(b, Quantity):
        if not b._meta:
            return r
        return isne
    return r


def ordering(f: operators.Operator, a, b):
    """Compute the ordering operation f(a, b)."""
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        if a._meta == b._meta:
            return f(a._data, b._data)
        errmsg = _build_error_message(f, a, b, error='unequal')
        raise MetadataValueError(errmsg) from None
    errmsg = _build_error_message(f, a, b, error='non-empty')
    if isinstance(a, Quantity):
        if not a._meta:
            return f(a._data, b)
        raise MetadataTypeError(errmsg) from None
    if isinstance(b, Quantity):
        if not b._meta:
            return f(a, b._data)
        raise MetadataTypeError(errmsg) from None
    return f(a, b)


def additive(f: operators.Operator, a, b):
    """Compute the additive operation f(a, b)."""
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        if a._meta == b._meta:
            return type(a)(f(a._data, b._data), **a._meta)
        errmsg = _build_error_message(f, a, b, error='unequal')
        raise MetadataValueError(errmsg) from None
    errmsg = _build_error_message(f, a, b, error='non-empty')
    if isinstance(a, Quantity):
        if not a._meta:
            return type(a)(f(a._data, b))
        raise MetadataTypeError(errmsg) from None
    if isinstance(b, Quantity):
        if not b._meta:
            return type(b)(f(a, b._data))
        raise MetadataTypeError(errmsg) from None
    return f(a, b)


def multiplicative(f: operators.Operator, a, b):
    """Compute the multiplicative operation f(a, b)."""
    if isinstance(a, Quantity) and isinstance(b, Quantity):
        keys = set(a._meta) & set(b._meta)
        meta = {}
        for key in keys:
            try:
                v = f(a._meta[key], b._meta[key])
            except TypeError as exc:
                errmsg = _build_error_message(f, a, b, error='type', key=key)
                raise MetadataTypeError(errmsg) from exc
            else:
                meta[key] = v
        for key, value in a._meta.items():
            if key not in keys:
                meta[key] = value
        for key, value in b._meta.items():
            if key not in keys:
                meta[key] = value
        return type(a)(f(a._data, b._data), **meta)
    if isinstance(a, Quantity):
        meta = {}
        for key, value in a._meta.items():
            try:
                v = f(value, b)
            except TypeError as exc:
                errmsg = _build_error_message(f, a, b, error='type', key=key)
                raise MetadataTypeError(errmsg) from exc
            else:
                meta[key] = v
        return type(a)(f(a._data, b), **meta)
    if isinstance(b, Quantity):
        meta = {}
        for key, value in b._meta.items():
            try:
                v = f(a, value)
            except TypeError as exc:
                errmsg = _build_error_message(f, a, b, error='type', key=key)
                raise MetadataTypeError(errmsg) from exc
            else:
                meta[key] = v
        return type(b)(f(a, b._data), **meta)
    return f(a, b)





import numerical

from . import methods
from . import _typeface


T = _typeface.TypeVar('T')

class Attribute(numerical.Object[T], numerical.mixins.NumpyMixin):
    """A metadata attribute with support for numerical operations.

    This class implements all operations required by the real-valued protocol
    (cf. `numerical.Real`). It is meant to serve as a base class from which to
    create metadata attributes for use in subclasses of `~Operand`.

    Notes
    -----
    Although this class implements all real-valued operations, it does not
    assume that its data is real-valued. In fact, it does not even assume that
    its data is numerical.
    """

    def __abs__(self):
        """Called for abs(self)."""
        return self

    def __pos__(self):
        """Called for +self."""
        return self

    def __neg__(self):
        """Called for -self."""
        return self

    def __eq__(self, other):
        """Called for self == other."""
        return equal(self, other)

    def __ne__(self, other):
        """Called for self != other."""
        return not equal(self, other)

    def __lt__(self, other):
        return NotImplemented

    def __le__(self, other):
        return NotImplemented

    def __gt__(self, other):
        return NotImplemented

    def __ge__(self, other):
        return NotImplemented

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __pow__(self, other):
        return self

    def __rpow__(self, other):
        return self


def equal(a, b):
    """Compute a == b between instances of `~Attribute`."""
    if isinstance(a, Attribute) and isinstance(b, Attribute):
        r = a._data == b._data
        try:
            iter(r)
        except TypeError:
            return bool(r)
        return all(r)
    return False




import numbers

from numerical import operators

from ._operations import (
    unary,
    equality,
    ordering,
    additive,
    multiplicative,
)


def __abs__(a):
    """Called for abs(a)."""
    return unary(operators.abs, a)

def __pos__(a):
    """Called for +a."""
    return unary(operators.pos, a)

def __neg__(a):
    """Called for -a."""
    return unary(operators.neg, a)

def __eq__(a, b):
    """Called for a == b."""
    return equality(operators.eq, a, b)

def __ne__(a, b):
    """Called for a != b."""
    return equality(operators.ne, a, b)

def __lt__(a, b):
    """Called for a < b."""
    return ordering(operators.lt, a, b)

def __le__(a, b):
    """Called for a <= b."""
    return ordering(operators.le, a, b)

def __gt__(a, b):
    """Called for a > b."""
    return ordering(operators.gt, a, b)

def __ge__(a, b):
    """Called for a >= b."""
    return ordering(operators.ge, a, b)

def __add__(a, b):
    """Called for a + b."""
    return additive(operators.add, a, b)

def __radd__(a, b):
    """Called for b + a."""
    return additive(operators.add, b, a)

def __sub__(a, b):
    """Called for a - b."""
    return additive(operators.sub, a, b)

def __rsub__(a, b):
    """Called for b - a."""
    return additive(operators.sub, b, a)

def __mul__(a, b):
    """Called for a * b."""
    return multiplicative(operators.mul, a, b)

def __rmul__(a, b):
    """Called for b * a."""
    return multiplicative(operators.mul, b, a)

def __truediv__(a, b):
    """Called for a / b."""
    return multiplicative(operators.truediv, a, b)

def __rtruediv__(a, b):
    """Called for b / a."""
    return multiplicative(operators.truediv, b, a)

def __floordiv__(a, b):
    """Called for a // b."""
    return multiplicative(operators.floordiv, a, b)

def __rfloordiv__(a, b):
    """Called for b // a."""
    return multiplicative(operators.floordiv, b, a)

def __mod__(a, b):
    """Called for a % b."""
    return multiplicative(operators.mod, a, b)

def __rmod__(a, b):
    """Called for b % a."""
    return multiplicative(operators.mod, b, a)

def __pow__(a, b):
    """Called for a ** b."""
    if isinstance(b, numbers.Real):
        return multiplicative(operators.pow, a, b)
    return NotImplemented

def __rpow__(a, b):
    """Called for b ** a."""
    return NotImplemented




import numpy

from . import _abstract
from . import methods
from . import mixins
from ._operations import equality
from . import _typeface


T = _typeface.TypeVar('T')

class Operand(_abstract.Object[T], mixins.NumpyMixin):
    """A concrete implementation of a real-valued object."""

    __abs__ = methods.__abs__
    __pos__ = methods.__pos__
    __neg__ = methods.__neg__

    __eq__ = methods.__eq__
    __ne__ = methods.__ne__
    __lt__ = methods.__lt__
    __le__ = methods.__le__
    __gt__ = methods.__gt__
    __ge__ = methods.__ge__

    __add__ = methods.__add__
    __radd__ = methods.__radd__
    __sub__ = methods.__sub__
    __rsub__ = methods.__rsub__
    __mul__ = methods.__mul__
    __rmul__ = methods.__rmul__
    __truediv__ = methods.__truediv__
    __rtruediv__ = methods.__rtruediv__
    __floordiv__ = methods.__floordiv__
    __rfloordiv__ = methods.__rfloordiv__
    __mod__ = methods.__mod__
    __rmod__ = methods.__rmod__
    __pow__ = methods.__pow__
    __rpow__ = methods.__rpow__

    def __array__(self, *args, **kwargs):
        """Called for numpy.array(self)."""
        return numpy.array(self._data, *args, **kwargs)

    def _apply_ufunc(self, ufunc, method, *args, **kwargs):
        if ufunc in (numpy.equal, numpy.not_equal):
            # NOTE: We are probably here because the left operand is a
            # `numpy.ndarray`, which would otherwise take control and return the
            # pure `numpy` result.
            f = getattr(ufunc, method)
            return equality(f, *args)
        data, meta = super()._apply_ufunc(ufunc, method, *args, **kwargs)
        return self._from_numpy(data, **meta)

    def _apply_function(self, func, types, args, kwargs):
        data, meta = super()._apply_function(func, types, args, kwargs)
        if data is NotImplemented:
            return data
        return self._from_numpy(data, **meta)

    def _get_numpy_array(self):
        return numpy.array(self._data)

    def _from_numpy(self, data, **meta):
        """Create a new instance after applying a numpy function."""
        if isinstance(data, (list, tuple)):
            r = [self._factory(array, **meta) for array in data]
            if isinstance(data, tuple):
                return tuple(r)
            return r
        return self._factory(data, **meta)

    def _factory(self, data, **meta):
        """Create a new instance from data and metadata.

        The default implementation uses the standard `__new__` constructor.
        Subclasses may overload this method to use a different constructor
        (e.g., a module-defined factory function).
        """
        return type(self)(data, **meta)


@Operand.implementation(numpy.array_equal)
def array_equal(
    x: numpy.typing.ArrayLike,
    y: numpy.typing.ArrayLike,
    **kwargs
) -> bool:
    """Called for numpy.array_equal(x, y)"""
    return numpy.array_equal(numpy.array(x), numpy.array(y), **kwargs)





import collections.abc

import numerical

from ._abstract import Quantity
from ._exceptions import OperationError
from . import _typeface


T = _typeface.TypeVar('T')

class Real:
    """Mixin for adding basic real-valued operator support."""

    def __abs__(self):
        return self

    def __pos__(self):
        return self

    def __neg__(self):
        return self

    def __eq__(self, other):
        return False

    def __ne__(self, other):
        return not (self == other)

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return (self < other) and (self == other)

    def __gt__(self, other):
        return not (self <= other)

    def __ge__(self, other):
        return not (self < other)

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __pow__(self, other):
        return self

    def __rpow__(self, other):
        return self


UserFunction = collections.abc.Callable[..., T]

class NumpyMixin(numerical.mixins.NumpyMixin):
    """Mixin for adding `numpy` support to objects with metadata.

    Notes
    -----
    - This class extends `numerical.mixins.NumpyMixin`. See that class for
      further documentation.
    """

    def _apply_ufunc(self, ufunc, method, *args, **kwargs):
        data = super()._apply_ufunc(ufunc, method, *args, **kwargs)
        f = getattr(ufunc, method)
        types = [type(arg) for arg in args]
        try:
            meta = self._apply_operator_to_metadata(f, *args, **kwargs)
        except OperationError as err:
            raise TypeError(
                f"Cannot apply numpy.{ufunc} to arguments with type(s) {types}"
            ) from err
        if method != 'at':
            return data, meta

    def _apply_function(self, func, types, args, kwargs):
        data = super()._apply_function(func, types, args, kwargs)
        try:
            meta = self._apply_operator_to_metadata(func, *args, **kwargs)
        except OperationError as err:
            raise TypeError(
                f"Cannot apply numpy.{func} to arguments with type(s) {types}"
            ) from err
        return data, meta

    def _apply_operator_to_metadata(self, f, *args, **kwargs):
        """Apply a numpy universal or public function to arguments."""
        keys = {
            k
            for x in args
            if isinstance(x, Quantity)
            for k in x._meta.keys()
        }
        meta = {}
        for key in keys:
            values = [
                x._meta[key]
                if isinstance(x, Quantity) else x
                for x in args
            ]
            errmsg = f"Attribute {key!r} does not support this operation"
            try:
                result = f(*values, **kwargs)
            except TypeError as err:
                if len(values) > 1 and any(v != values[0] for v in values):
                    raise TypeError(errmsg) from err
                else:
                    meta[key] = values[0]
            except OperationError as err:
                raise OperationError(errmsg) from err
            else:
                meta[key] = result
        return meta.copy()

